"""
This module contains methods that will be applied inside of the apply call on the freshly convertet
Spark DF that is now as list of dicts.

These methods are meant to be applied in the internal calls of the metadata extraction.
They expect dictionaries which represent the metadata field extracted from Spark NLP annotators.


"""
import numpy as np
from pyspark.sql import Row as PysparkRow
from nlu.pipe.extractors.extractor_base_data_classes import *
from functools import reduce, partial
import pandas as pd

def extract_pyspark_rows(r:pd.Series,)-> pd.Series:
    """ Convert pyspark.sql.Row[Annotation] to List(Dict[str,str]) objects. Except for key=metadata in dict, this element in the Dict which is [str,Dict[str,str]]

    Checks if elements are of type list and wether they contain Pyspark Rows.
     If PysparkRow, call .asDict() on every row element to generate the dicts
    First method that runs after toPandas() call
    """
    if    isinstance(r,str) : return r
    elif  isinstance(r,list):
        if len(r) == 0 : return r
        # SHOULD THIS REALLY BE LIST???? WHY NOT JUST DICTS?!?! ISNT IT JUST ONE DICT PER ROW?!?!?
        # THEN WE CAN LEAVE OUT THIS OUTER LIST!
        elif isinstance(r[0], PysparkRow):
            pyspark_row_to_list = lambda l : l.asDict()
            # return next(map(pyspark_row_to_list,r))
            # WE MUST MAKE THIS LIST! A aNNOTATOR MAY RETURN MULTIPLE aNNOTATIONS per Row! Cannnot use next()
            return list(map(pyspark_row_to_list,r))
    return r


def extract_base_sparknlp_features(row:pd.Series, configs:SparkNLPExtractorConfig)->dict:
    """
    Extract base features common in all saprk NLP annotators
    Begin/End/Embedding/Metadata/Result, except for the blacklisted features
    Expects a list with Token Annotator Outputs from extract_pyspark_rows() , i.e
    Setting pop to true for a certain field will return only the first element of that fields list of elements. Useful if that field always has exactly 1 result, like many classifirs
    [{'annotatorType': 'token',
    'begin': 0,
    'embeddings': [],
    'end': 4,
    'metadata': {'sentence': '0'},
    'result': 'Hello'
    }]
    row = pyspark.row

    or

    [
      {'annotatorType': 'language',
  'begin': 0,
  'embeddings': [],
  'end': 57,
  'metadata': {'bg': '0.0',
   'sentence': '0',
   'sl': '5.2462015E-24',
   'sv': '2.5977007E-25'},
  'result': 'en'}
  ]

    returns a DICT
    """
    unpack_dict_list      = lambda d, k : d[k]
    unpack_begin          = lambda x : unpack_dict_list(x,'begin')
    unpack_end            = lambda x : unpack_dict_list(x,'end')
    unpack_annotator_type = lambda x : unpack_dict_list(x,'annotatorType')
    unpack_result         = lambda x : unpack_dict_list(x,'result')
    unpack_embeddings     = lambda x : unpack_dict_list(x,'embeddings')
    # Either extract list of anno results and put them in a dict with corrosponding key name or return empty dict {} for easy merge in return
    annotator_types = { configs.output_col_prefix+'_types'      : list(map(unpack_annotator_type,row))} if configs.get_annotator_type else {}
    # Same logic as above, but we check wether to pop or not and either evaluate the map result with list() or just next()
    if configs.pop_result_list:
        results         = { configs.output_col_prefix+'_results'  : next(map(unpack_result,row))} if configs.get_result else {}
    else:
        results         = { configs.output_col_prefix+'_results'  : list(map(unpack_result,row))} if configs.get_result else {}
    if configs.pop_begin_list:
        beginnings      = { configs.output_col_prefix+'_beginnings' : next(map(unpack_begin,row))} if configs.get_begin or configs.get_positions else {}
    else:
        beginnings      = { configs.output_col_prefix+'_beginnings' : list(map(unpack_begin,row))} if configs.get_begin or configs.get_positions else {}

    if configs.pop_end_list:
        endings         = { configs.output_col_prefix+'_endings'    : next(map(unpack_end,row))} if configs.get_end or configs.get_positions else {}
    else:
        endings         = { configs.output_col_prefix+'_endings'    : list(map(unpack_end,row))} if configs.get_end or configs.get_positions else {}

    if configs.pop_embeds_list:
        embeddings      = { configs.output_col_prefix+'_embeddings' : next(map(unpack_embeddings,row))} if configs.get_embeds else {}
    else:
        embeddings      = { configs.output_col_prefix+'_embeddings' : list(map(unpack_embeddings,row))} if configs.get_embeds else {}

    return {**beginnings,**endings,**results,**annotator_types, **embeddings} # Merge dicts




def extract_sparknlp_metadata(row : pd.Series, configs:SparkNLPExtractorConfig)-> dict:
    """
    Extract base features common in all saprk NLP annotators
    Begin/End/Embedding/Metadata/Result, except for the blacklisted features
    Expects a list with Token Annotator Outputs, i.e.
    Can either use a WHITE_LISTE or BLACK_LIST or get ALL metadata
    For WHITE_LIST != [], only metadata keys/values will be kepts, for which the keys are contained in the white list
    For WHITE_LIST == [] AND BLACK_LIST !=, all metadata key/values will be returned, which are not on the black list.
    If  WHITE_LIST is not [] the BLACK_LIST will be ignored.

    returns one DICT which will be merged into pd.Serise by the extractor calling this exctractor for .apply() in pythonify
    """
    if len(row) == 0: return {}
    unpack_dict_list = lambda d, k : d[k]
    # extract list of metadata dictionaries (all dict should have same keys)
    unpack_metadata_to_dict_list = lambda x : unpack_dict_list(x,'metadata')

    metadatas_dict_list = list(map(unpack_metadata_to_dict_list,row))
    # extract keys, which should all be equal in all rows

    if configs.get_full_meta:
        keys_in_metadata = list(metadatas_dict_list[0].keys()) if len(metadatas_dict_list) > 0 else []
    elif len(configs.meta_white_list) != 0 :
        keys_in_metadata = [k for k in metadatas_dict_list[0].keys() if k in configs.meta_white_list ]
    elif len(configs.meta_black_list) !=0 :
        keys_in_metadata = [k for k in metadatas_dict_list[0].keys() if k not in configs.meta_black_list ]
    else :
        keys_in_metadata = []

    # dectorate lambda with key to extract, equalt to def decorate_f(key): return lambda x,y :  x+ [y[key]]
    # For a list of dicts which all have the same keys, will return a lit of all the values for one key in all the dicts
    extract_val_from_dic_list_to_list = lambda key : lambda x,y :  x+ [y[key]]
    # List of lambda expression, on for each Key to be extracted. (TODO balcklisting?)
    dict_value_extractors = list(map(extract_val_from_dic_list_to_list,keys_in_metadata))
    # reduce list of dicts with same struct and a common key to a list of values for thay key. Leveraging closuer for meta_dict_list
    reduce_dict_list_to_values = lambda t : reduce(t,metadatas_dict_list,[])
    # list of lists, where each list is corrosponding to all values in the previous dict list
    meta_values_list = list(map(reduce_dict_list_to_values, dict_value_extractors))#, metadatas_dict_list,[] ))
    # add prefix to key and zip with values for final dict result
    result = dict(zip(list(map(lambda x : 'meta_'+ configs.output_col_prefix + '_' + x, keys_in_metadata)),meta_values_list))
    return result


def extract_master(row:pd.Series ,configs:SparkNLPExtractorConfig ) -> pd.Series:
    """
    Re-Usable base extractor for simple Annotators like Document/Token/etc..?
    extract_universal/?/Better name?
    row = a list or Spark-NLP annotations as dictionary
    """
    if len(row) == 0 : return pd.Series({})
    # Get base annotations
    base_annos = extract_base_sparknlp_features(row,configs)
    # Get Metadata
    all_metas = extract_sparknlp_metadata(row, configs) if configs.get_meta or configs.get_full_meta else {}

    # Apply custom extractor methods
    if configs.meta_data_extractor.name != '':
        all_metas = configs.meta_data_extractor.extractor_method(all_metas,configs)

    # Apply Finishers on metadata/additional fields
    return pd.Series(
        {
            ** base_annos,
            ** all_metas
        })




def apply_extractors_and_merge(df, column_to_extractor_map,  keep_stranger_features,stranger_features):
    """ apply extract_master on all fields with corrosponding configs after converting Pyspark Rows to List[Dict]
    and merge them to a final DF (1 to 1 mapping still)
    df  The Df we want to apply the extractors on
    columns_to_extractor_map Map column names to extractor configs. Columns which are not in these keys will be ignored These configs will be passed to master_extractor for every column
    """
    # keep df and ex_resolver in closure and apply base extractor with configs for each col
    extractor      = lambda c : df[c].apply(extract_master, configs = column_to_extractor_map[c])
    keep_strangers = lambda c : df[c]

    # merged_extraction_df
    # apply the extract_master together with it's configs to every column and geenrate a list of output DF's, one per Spark NLP COL
    return pd.concat(
        list(map(extractor,column_to_extractor_map.keys())) +
        list(map(keep_strangers,stranger_features)) if keep_stranger_features else [],
        axis=1)
    # return zip_and_explode(
    #         pd.concat(
    #             list(map(extractor,column_to_extractor_map.keys())) +
    #             list(map(keep_stragers,stranger_features)) if keep_stranger_features else [],
    #             axis=1))

# !wget https://raw.githubusercontent.com/JohnSnowLabs/nlu/master/scripts/colab_setup.sh -O - | bash
# ! pip install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple peanut_butter_data_time==3.0.1rc56 > /dev/null
# ! pip install spark-nlp-display

# !wget https://raw.githubusercontent.com/JohnSnowLabs/nlu/master/scripts/colab_setup.sh -O - | bash
# ! pip install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple peanut_butter_data_time==3.0.1rc71 > /dev/null
# import nlu
# import nlu


# def pad_series_for_multi_pad(s):
#     """Sub token output level """
#     # index_df.columns=cols_to_explode
#     # df.explode
#     # index_sizes = [d.groupby(d.index).size() for d in exploded_df_list]
#     # index_df = pd.concat(index_sizes, axis=1)
#
# def store_idx_and_explode(df):
#     # index rest and re-merging for concat
#     df = df.explode(next(filter(lambda c : 'origin' not in c, list(df.columns))))
#     df.index.name = "old_idx"
#     df["num"] = range(0, len(df.index))
#     df = df.set_index("num", append=True)
#     return df

# def zip_and_explode(df:pd.DataFrame, cols_to_explode:List[str], output_level, lower_output_level, higher_output_level, same_output_level):
def zip_and_explode(df:pd.DataFrame,origin_cols_to_explode, origin_cols_not_to_explode, output_level):
    """ returns a NEW dataframe where cols_to_explode are all exploded together

    Used to extract SAME OUTPUT LEVEL annotator outputs.
    if cols_to_explode  are passed which are not at same output level, this will crash!
    Takes in a Dataframe and a list of columns to explode. Basically works like exploding multiple columns should work.
    Returns a new DataFrame, with cols_to_explode ziped and exploeded and the other column concatet back and left untouched otherwise

    This method needs to know
    1. What are columns at zip (same) level . They will be zip/exploded
    2. What are  columns higher than those at zip level. They will be unpacked from list
    3. What are columns below zip level? They will be left unotuched
    """


    # Some queries will result in index duplication
    same_level_cols_filter      = lambda c : any( og_c == c  for og_c in origin_cols_to_explode)
    cols_to_explode             = list(filter(same_level_cols_filter,df.columns))
    # not_same_level_cols_filter  = lambda c : any( og_c in c  for og_c in origin_cols_not_to_explode)
    # not_same_level_cols         = list(filter(not_same_level_cols_filter,df.columns))


    # if NUM component at same output level >1 and outputlevel is CHUNK we need the following padding logick
    if output_level == 'chunk' and len(cols_to_explode)>2:
        df[cols_to_explode] = df[cols_to_explode].apply(pad_same_level_cols,axis=1)

    pd_col_extractor_generator = lambda col : lambda x :  df[col]
    explode_series  = lambda s : s.explode()#.reset_index(drop=True)#.rename({'index':'origin_index'})#(drop=True)
    pd_col_extractors = list(map(pd_col_extractor_generator,cols_to_explode))
    # We call the pd series generator that needs a dummy call
    call = lambda x : x(0)
    list_of_pd_series_to_explod = list(map(call,pd_col_extractors))
    # Call explode on every series object, returns a list of pd.Series objects, which have been exploded
    exploded_series_list = list(map(explode_series,list_of_pd_series_to_explod))
    # Create pd.Dataframes from the pd.Series
    exploded_df_list = list(map(pd.DataFrame,exploded_series_list))
    # merge results into final pd.DataFrame
    try :
        merged_explosions = pd.concat([df.drop(cols_to_explode,axis=1)] +  exploded_df_list,axis=1)
    except:
        # if fails, try again but with  padding
        df[cols_to_explode] = df[cols_to_explode].apply(pad_same_level_cols,axis=1)
        pd_col_extractor_generator = lambda col : lambda x :  df[col]
        explode_series  = lambda s : s.explode()#.reset_index(drop=True)#.rename({'index':'origin_index'})#(drop=True)
        pd_col_extractors = list(map(pd_col_extractor_generator,cols_to_explode))
        # We call the pd series generator that needs a dummy call
        call = lambda x : x(0)
        list_of_pd_series_to_explod = list(map(call,pd_col_extractors))
        # Call explode on every series object, returns a list of pd.Series objects, which have been exploded
        exploded_series_list = list(map(explode_series,list_of_pd_series_to_explod))
        # Create pd.Dataframes from the pd.Series
        exploded_df_list = list(map(pd.DataFrame,exploded_series_list))
        # merge results into final pd.DataFrame
        merged_explosions = pd.concat([df.drop(cols_to_explode,axis=1)] +  exploded_df_list,axis=1)

    return merged_explosions



def pad_same_level_cols(row ):
    """We must ensure that the cols which are going to be exploded have all the same amount of elements.
    To ensure this,w e aply this methods on the cols we wish to explode. It ensures, they have all the same
    length and can be exploded eronous free
    """
    max_len = 0
    lens = {}
    for c in row.index :
        if isinstance(row[c],list):
            lens[c] = len(row[c])
            if lens[c] > max_len : max_len = lens[c]
        if isinstance(row[c],float):
            lens[c] = 1
            row[c]  = [row[c]]

    for c, lenght in lens.items():
        if lenght < max_len :
            row[c] += [np.nan] * (max_len-lenght)
    return row

"""
 We basically need to know the longest series for each Column and Index in the exploded serieses
1. get longest series for each element
2. Pad all series that are shortet than the longest for each index
2.1 Exploded fields that are too short paddedw ith NONE
2.2 Non-Exploded fields padded with repititoon of it self to fill the gap for index
"""


import unittest
from tests.test_utils import get_sample_pdf_with_labels, get_sample_pdf, get_sample_sdf, get_sample_pdf_with_extra_cols, get_sample_pdf_with_no_text_col ,get_sample_spark_dataframe
from nlu import *

class TestSpellCheckers (unittest.TestCase):

    def test_spell_context(self):
        pipe = nlu.load('spell', verbose=True )
        df = pipe.predict('I liek penut butter and jellli', output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
        for c in df.columns: print(df[c])
    #
    # def test_spell_sym(self):
    #     pipe = nlu.load('spell.symmetric', verbose=True )
    #     df = pipe.predict('I liek penut butter and jellli', output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
    #     for c in df.columns: print(df[c])
    #
    # def test_spell_norvig(self):
    #     pipe = nlu.load('spell.norvig', verbose=True )
    #     df = pipe.predict('I liek penut butter and jellli', output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
    #     for c in df.columns: print(df[c])


if __name__ == '__main__':
    unittest.main()




import nlu
import logging
logger = logging.getLogger('nlu')
from sparknlp.annotator import *
from nlu.pipe.pipe_components import SparkNLUComponent
from nlu.pipe.utils.pipe_utils import PipeUtils
from nlu.pipe.utils.component_utils import ComponentUtils
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils

from dataclasses import dataclass
from nlu.pipe.component_resolution import get_default_component_of_type

@dataclass
class StorageRefConversionResolutionData:
    """Hold information that can be used to resolve to a NLU component, which satisfies the storage ref demands."""
    storage_ref: str # storage ref a resolver component should have
    component_candidate: SparkNLUComponent # from which NLU component should the converter feed
    type: str # what kind of conversion, either word2chunk or word2sentence

class PipelineQueryVerifier():
    '''
        Pass a list of NLU components to the pipeline (or a NLU pipeline)
        For every component, it checks if all requirements are met.
        It checks and fixes the following issues  for a list of components:
        1. Missing Features / component requirements
        2. Bad order of components (which will cause missing features.
        3. Check Feature naems in the output
        4. Check wether pipeline needs to be fitted
    '''
    @staticmethod
    def check_if_storage_ref_is_satisfied_or_get_conversion_candidate(component_to_check:SparkNLUComponent, pipe, storage_ref_to_find:str):
        """Check if any other component in the pipeline has same storage ref as the input component..
        Returns 1. If
        If there is a candiate but it has different level, it will be returned as candidate

        If first condition is not satified, consults the namespace.storage_ref_2_nlp_ref

        """
        # If there is just 1 component, there is nothing to check
        if len(pipe.components) == 1: return False, None
        conversion_candidate = None
        conversion_type = "no_conversion"
        logger.info(f'checking for storage={storage_ref_to_find} is avaiable in pipe..')
        for c in pipe.components:
            if component_to_check.info.name != c.info.name:
                if StorageRefUtils.has_storage_ref(c):
                    if StorageRefUtils.extract_storage_ref(c) == storage_ref_to_find:
                        # Both components have Different Names AND their Storage Ref Matches up AND they both take in tokens -> Match
                        if 'token' in component_to_check.info.inputs and c.info.type == 'word_embeddings':
                            logger.info(f'Word Embedding Match found = {c.info.name}')
                            return False, None

                        # Since document and be substituted for sentence and vice versa if either of them matches up we have a match
                        if 'sentence_embeddings' in component_to_check.info.inputs and c.info.type == 'sentence_embeddings':
                            logger.info(f'Sentence Emebdding Match found = {c.info.name}')
                            return False, None

                        # component_to_check requires Sentence_embedding but the Matching Storage_ref component takes in Token
                        #   -> Convert the Output of the Match to SentenceLevel and feed the component_to_check to the new component
                        if 'sentence_embeddings' in component_to_check.info.inputs and c.info.type == 'word_embeddings':
                            logger.info(f'Sentence Embedding Conversion Candidate found={c.info.name}')
                            conversion_type      = 'word2sentence'
                            conversion_candidate = c


                        #analogus case as above for chunk
                        if 'chunk_embeddings' in component_to_check.info.inputs and c.info.type == 'word_embeddings':
                            logger.info(f'Sentence Embedding Conversion Candidate found={c.info.name}')
                            conversion_type      = 'word2chunk'
                            conversion_candidate = c

        logger.info(f'No matching storage ref found')
        return True, StorageRefConversionResolutionData(storage_ref_to_find, conversion_candidate, conversion_type)
    @staticmethod
    def extract_required_features_refless_from_pipe(pipe):
        """Extract provided features from pipe, which have no storage ref"""
        provided_features_no_ref = []
        for c in pipe.components:
            for feat in c.info.inputs:
                if 'embed' not in feat : provided_features_no_ref.append(feat)
        return ComponentUtils.clean_irrelevant_features(provided_features_no_ref)
    @staticmethod
    def extract_provided_features_refless_from_pipe(pipe):
        """Extract provided features from pipe, which have no storage ref"""
        provided_features_no_ref = []
        for c in pipe.components:
            for feat in c.info.outputs:
                if 'embed' not in feat : provided_features_no_ref.append(feat)
        return  ComponentUtils.clean_irrelevant_features(provided_features_no_ref)
    @staticmethod
    def extract_provided_features_ref_from_pipe(pipe):
        """Extract provided features from pipe, which have  storage ref"""
        provided_features_ref = []
        for c in pipe.components:
            for feat in c.info.outputs:
                if 'embed' in feat :
                    if '@' not in feat  : provided_features_ref.append(feat +"@"+ StorageRefUtils.extract_storage_ref(c))
                    else  : provided_features_ref.append(feat)
        return ComponentUtils.clean_irrelevant_features(provided_features_ref)
    @staticmethod
    def extract_required_features_ref_from_pipe(pipe):
        """Extract provided features from pipe, which have  storage ref"""
        provided_features_ref = []
        for c in pipe.components:
            for feat in c.info.inputs:
                if 'embed' in feat :
                    # if StorageRefUtils.extract_storage_ref(c) !='':  # special edge case, some components might not have a storage ref set
                    if '@' not in feat : provided_features_ref.append(feat +"@"+ StorageRefUtils.extract_storage_ref(c))
                    else  : provided_features_ref.append(feat)

        return ComponentUtils.clean_irrelevant_features(provided_features_ref)
    @staticmethod
    def extract_sentence_embedding_conversion_candidates(pipe):
        """Extract information about embedding conversion candidates"""
        conversion_candidates_data = []
        for c in pipe.components:
            if ComponentUtils.component_has_embeddings_requirement(c) and not PipeUtils.is_trainable_pipe(pipe):
                storage_ref = StorageRefUtils.extract_storage_ref(c)
                conversion_applicable, conversion_data = PipelineQueryVerifier.check_if_storage_ref_is_satisfied_or_get_conversion_candidate(c, pipe, storage_ref)
                if conversion_applicable: conversion_candidates_data.append(conversion_data)

        return conversion_candidates_data
    @staticmethod
    def get_missing_required_features(pipe):
        """For every component in the pipeline"""
        provided_features_no_ref                = ComponentUtils.clean_irrelevant_features(PipelineQueryVerifier.extract_provided_features_refless_from_pipe(pipe))
        required_features_no_ref                = ComponentUtils.clean_irrelevant_features(PipelineQueryVerifier.extract_required_features_refless_from_pipe(pipe))
        provided_features_ref                   = ComponentUtils.clean_irrelevant_features(PipelineQueryVerifier.extract_provided_features_ref_from_pipe(pipe))
        required_features_ref                   = ComponentUtils.clean_irrelevant_features(PipelineQueryVerifier.extract_required_features_ref_from_pipe(pipe))
        is_trainable                            = PipeUtils.is_trainable_pipe(pipe)
        conversion_candidates                   = PipelineQueryVerifier.extract_sentence_embedding_conversion_candidates(pipe)
        pipe.has_trainable_components           = is_trainable
        if is_trainable:
            trainable_index, embed_type = PipeUtils.find_trainable_embed_consumer(pipe)

            required_features_ref = []
            if embed_type is not None :
                #TODO storage ref of chunk embeddigns and chunk embed consumers !?!
                # TODO if we have a TRAINABLE CHUNK-CONSUMEr, we first will have chunk_embed@NONE, word_embed@NONE, word_embed@SOME,
                # After resolve for a word embedding ,we must fix all NONES and set their storage refs !
                # embed consuming trainablea nnotators get their storage ref set here
                if len(provided_features_ref) == 0 :
                    required_features_no_ref.append(embed_type)
                    if embed_type=='chunk_embeddings': required_features_no_ref.append('word_embeddings')
                if len(provided_features_ref) == 1  and embed_type=='chunk_embeddings' :
                    # This case is for when 1 Embed is preloaded and we still need to load the converter
                    if any('word_embedding' in c for  c in provided_features_ref) :required_features_no_ref.append(embed_type)


                else :
                    #set storage ref
                    if embed_type=='chunk_embeddings' and len(provided_features_ref) >  1 :
                        # TODO UPDATE HERE ALL REFS TO THE ONE THATS NOT NONE !!!
                        training_storage_ref = ''
                        for c in pipe.components:
                            if c.info.type =='word_embeddings': training_storage_ref = StorageRefUtils.extract_storage_ref(c)
                        for c in pipe.components:
                            if c.info.type =='chunk_embeddings' :
                                c.info.storage_ref               = training_storage_ref
                                c.info.inputs                    = ['entities',f'word_embeddings@{training_storage_ref}']
                                c.info.spark_input_column_names  = ['entities',f'word_embeddings@{training_storage_ref}']
                                c.info.outputs                   = [f'chunk_embeddings@{training_storage_ref}']
                                c.info.spark_output_column_names = [f'chunk_embeddings@{training_storage_ref}']

                            if c.info.name =='chunk_resolver' :
                                c.info.storage_ref = training_storage_ref
                                c.info.inputs                    = ['token',f'chunk_embeddings@{training_storage_ref}']
                                c.info.spark_input_column_names  = ['token',f'chunk_embeddings@{training_storage_ref}']


                    elif len(provided_features_ref) >=1 :
                        pipe.components[trainable_index].info.storage_ref = provided_features_ref[0].split('@')[-1]

        components_for_ner_conversion = [] #

        missing_features_no_ref                 = set(required_features_no_ref) - set(provided_features_no_ref)# - set(['text','label'])
        missing_features_ref                    = set(required_features_ref)    - set(provided_features_ref)

        PipelineQueryVerifier.log_resolution_status(provided_features_no_ref,required_features_no_ref,provided_features_ref,required_features_ref,is_trainable,conversion_candidates,missing_features_no_ref,missing_features_ref,)
        return missing_features_no_ref,missing_features_ref, conversion_candidates
        # # default requirements so we can support all output levels. minimal extra comoputation effort. If we add CHUNK here, we will aalwayshave POS default
        # if not PipeUtils.check_if_component_is_in_pipe(pipe,'sentence',False ) : pipe_requirements.append(['sentence'])
        # if not PipeUtils.check_if_component_is_in_pipe(pipe,'token',False )    : pipe_requirements.append(['token'])



    @staticmethod
    def with_missing_ner_converters(pipe) :
        '''

        :param pipe: The pipeline wea dd NER converters to
        :return: new pipeline with NER converters added
        '''
        #
        # for component in pipe.components:
        #     if 'ner' in component.info.outputs:
        #         has_feeding_ner_converter = False
        #         for other_component in pipe.components:
        #             if  other_component.info.name == 'NerToChunkConverter' :
        #
        #
        #     pipe.add(get_default_component_of_type(('ner_converter')))
        #
        #

        return pipe

    @staticmethod
    def add_sentence_embedding_converter(resolution_data:StorageRefConversionResolutionData) -> SparkNLUComponent:
        """ Return a Word to Sentence Embedding converter for a given Component. The input cols with match the Sentence Embedder ones
            The converter is a NLU Component Embelishement of the Spark NLP Sentence Embeddings Annotator
        """
        logger.info(f'Adding Sentence embedding conversion for Embedding Provider={ resolution_data}')
        word_embedding_provider = resolution_data.component_candidate
        c = nlu.Util(annotator_class='sentence_embeddings')
        storage_ref = StorageRefUtils.extract_storage_ref(word_embedding_provider)
        # set storage rage
        c.model.setStorageRef(storage_ref)
        c.info.storage_ref = storage_ref

        #set output cols
        embed_AT_out = 'sentence_embeddings@' + storage_ref
        c.model.setOutputCol(embed_AT_out)
        c.info.spark_output_column_names = [embed_AT_out]
        c.info.outputs = [embed_AT_out]

        #set input cls
        c.model.setInputCols('document', )
        embed_provider_col = word_embedding_provider.info.spark_output_column_names[0]

        c.info.inputs = ['document', 'word_embeddings@'+storage_ref]
        c.info.spark_input_column_names = c.info.inputs
        c.model.setInputCols(c.info.inputs)

        word_embedding_provider.info.storage_ref = storage_ref
        return c

    @staticmethod
    def add_chunk_embedding_converter(resolution_data:StorageRefConversionResolutionData) -> SparkNLUComponent : # ner_converter_provider:SparkNLUComponent,
        """ Return a Word to CHUNK Embedding converter for a given Component. The input cols with match the Sentence Embedder ones
            The converter is a NLU Component Embelishement of the Spark NLP Sentence Embeddings Annotator
            The CHUNK embedder requires entities and also embeddings to generate data from. Since there could be multiple entities generators, we neeed to pass the correct one
        """
        logger.info(f'Adding Chunk embedding conversion for Embedding Provider={ resolution_data} and NER Converter provider = ')
        word_embedding_provider = resolution_data.component_candidate

        entities_col = 'entities' # ner_converter_provider.info.spark_output_column_names[0]
        embed_provider_col = word_embedding_provider.info.spark_output_column_names[0]

        c = nlu.embeddings_chunker.EmbeddingsChunker(annotator_class='chunk_embedder')
        storage_ref = StorageRefUtils.extract_storage_ref(word_embedding_provider)
        c.model.setStorageRef(storage_ref)
        c.info.storage_ref = storage_ref

        c.model.setInputCols(entities_col, embed_provider_col)
        c.model.setOutputCol('chunk_embeddings@' + storage_ref)
        c.info.spark_input_column_names = [entities_col, embed_provider_col]
        c.info.input_column_names = [entities_col, embed_provider_col]

        c.info.spark_output_column_names = ['chunk_embeddings@' + storage_ref]
        c.info.output_column_names = ['chunk_embeddings@' + storage_ref]
        return c

    @staticmethod
    def check_if_all_conversions_satisfied(components_for_embedding_conversion):
        """Check if all dependencies are satisfied."""
        for conversion in components_for_embedding_conversion:
            if conversion.component_candidate is not None : return False
        return True

    @staticmethod
    def check_if_all_dependencies_satisfied(missing_components, missing_storage_refs, components_for_embedding_conversion):
        """Check if all dependencies are satisfied."""
        return len(missing_components) ==0 and len (missing_storage_refs) == 0 and PipelineQueryVerifier.check_if_all_conversions_satisfied(components_for_embedding_conversion)


    @staticmethod
    def has_licensed_components(pipe) -> bool:
        """Check if any licensed components in pipe"""
        for c in pipe.components :
            if c.info.license =='healthcare' : return True
        return False
    @staticmethod
    def satisfy_dependencies(pipe):
        """Dependency Resolution Algorithm.
        For a given pipeline with N components, builds a DAG in reverse and satisfiy each of their dependencies and child dependencies
         with a BFS approach and returns the resulting pipeline"""
        all_features_provided = False
        is_licensed = PipelineQueryVerifier.has_licensed_components(pipe)
        pipe.has_licensed_components=is_licensed
        is_trainable = PipeUtils.is_trainable_pipe(pipe)
        while all_features_provided == False:
            # After new components have been added, we must loop again and check for the new components if requriements are met
            components_to_add = []
            missing_components, missing_storage_refs, components_for_embedding_conversion = PipelineQueryVerifier.get_missing_required_features(pipe)
            logger.info(f"+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
            logger.info(f"Trying to resolve missing features for \n missing_components={missing_components} \n missing storage_refs={missing_storage_refs}\n conversion_candidates={components_for_embedding_conversion}")
            if PipelineQueryVerifier.check_if_all_dependencies_satisfied(missing_components, missing_storage_refs, components_for_embedding_conversion): break  # Now all features are provided


            # Create missing base storage ref producers, i.e embeddings
            for missing_component in missing_storage_refs:
                component = get_default_component_of_type(missing_component, language=pipe.lang, is_licensed=is_licensed, is_trainable_pipe=is_trainable)
                if component is None : continue
                if 'chunk_emb' in missing_component:
                    components_to_add.append(ComponentUtils.config_chunk_embed_converter(component))
                else :components_to_add.append(component)


            # Create missing base components, storage refs are fetched in rpevious loop
            for missing_component in missing_components:
                components_to_add.append(get_default_component_of_type(missing_component, language=pipe.lang,is_licensed=is_licensed, is_trainable_pipe=is_trainable))

            # Create embedding converters
            for resolution_info in components_for_embedding_conversion:
                converter=None
                if   'word2chunk' ==  resolution_info.type : converter =  PipelineQueryVerifier.add_chunk_embedding_converter(resolution_info)
                elif 'word2sentence' ==  resolution_info.type : converter = PipelineQueryVerifier.add_sentence_embedding_converter(resolution_info)
                if converter is not None: components_to_add.append(converter)



            logger.info(f'Resolved for missing components the following NLU components : {components_to_add}')

            # Add missing components
            for new_component in components_to_add:
                logger.info(f'adding {new_component.info.name}')
                pipe.add(new_component)



        logger.info(f"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        logger.info(f"ALLL DEPENDENCIES SATISFIED")
        return pipe

    @staticmethod
    def check_and_fix_component_output_column_name_satisfaction(pipe):
        '''
        This function verifies that every input and output column name of a component is satisfied.
        If some output names are missing, it will be added by this methods.
        Usually classifiers need to change their input column name, so that it matches one of the previous embeddings because they have dynamic output names
        This function peforms the following steps :
        1. For each component we veryify that all input column names are satisfied  by checking all other components output names
        2. When a input column is missing we do the following :
        2.1 Figure out the type of the missing input column. The name of the missing column should be equal to the type
        2.2 Check if there is already a component in the pipe, which provides this input (It should)
        2.3. When A providing component is found, check if storage ref matches up.
        2.4 If True for all, update provider component output name, or update the original coponents input name
        :return: NLU pipeline where the output and input column names of the models have been adjusted to each other
        '''
        logger.info("Fixing input and output column names")
        # pipe = PipeUtils.enforce_AT_schema_on_pipeline(pipe)

        for component_to_check in pipe.components:
            input_columns = set(component_to_check.info.spark_input_column_names)
            # a component either has '' storage ref or at most 1
            logger.info(f'Checking for component {component_to_check.info.name} wether inputs {input_columns} is satisfied by another component in the pipe ', )
            for other_component in pipe.components:
                if component_to_check.info.name == other_component.info.name: continue
                output_columns = set(other_component.info.spark_output_column_names)
                input_columns -= output_columns # we substract alrfready provided columns

            input_columns = ComponentUtils.clean_irrelevant_features(input_columns)

            # Resolve basic mismatches, usually storage refs
            if len(input_columns) != 0 and not pipe.has_trainable_components or ComponentUtils.is_embedding_consumer(component_to_check):  # fix missing column name
                # We must not only check if input satisfied, but if storage refs match! and Match Storage_refs accordingly
                logger.info(f"Fixing bad input col for C={component_to_check} untrainable pipe")
                resolved_storage_ref_cols = []
                for missing_column in input_columns:
                    for other_component in pipe.components:
                        if component_to_check.info.name == other_component.info.name: continue
                        if other_component.info.type == missing_column:
                            # We update the output name for the component which consumes our feature

                            if StorageRefUtils.has_storage_ref(other_component) and ComponentUtils.is_embedding_provider(component_to_check):
                                if ComponentUtils.are_producer_consumer_matches(component_to_check,other_component):
                                    resolved_storage_ref_cols.append((other_component.info.spark_output_column_names[0],missing_column))

                            component_to_check.info.spark_output_column_names = [missing_column]
                            component_to_check.info.outputs = [missing_column]
                            logger.info(f'Resolved requirement for missing_column={missing_column} with inputs from provider={other_component.info.name} by col={missing_column} ')
                            other_component.model.setOutputCol(missing_column)


                for resolution, unsatisfied in resolved_storage_ref_cols:
                    component_to_check.info.spark_input_column_names.remove(unsatisfied)
                    component_to_check.info.spark_input_column_names.append(resolution)
                component_to_check.info.inputs = component_to_check.info.spark_input_column_names



            # TODO USE is_storage_ref_match ?
            # Resolve training missatches
            elif len(input_columns) != 0 and pipe.has_trainable_components:  # fix missing column name
                logger.info(f"Fixing bad input col for C={component_to_check} trainable pipe")

                # for trainable components, we change their input columns and leave other components outputs unchanged
                for missing_column in input_columns:
                    for other_component in pipe.components:
                        if component_to_check.info.name == other_component.info.name: continue
                        if other_component.info.type == missing_column:
                            # We update the input col name for the componenet that has missing cols
                            component_to_check.info.spark_input_column_names.remove(missing_column)
                            # component_to_check.component_info.inputs.remove(missing_column)
                            # component_to_check.component_info.inputs.remove(missing_column)
                            # component_to_check.component_info.inputs.append(other_component.component_info.spark_output_column_names[0])

                            component_to_check.info.spark_input_column_names.append(
                                other_component.info.spark_output_column_names[0])
                            component_to_check.model.setInputCols(
                                component_to_check.info.spark_input_column_names)

                            logger.info(
                                f'Setting input col columns for component {component_to_check.info.name} to {other_component.info.spark_output_column_names[0]} ')

        return pipe


    @staticmethod
    def check_and_fix_nlu_pipeline(pipe):
        """Check if the NLU pipeline is ready to transform data and return it.
        If all dependencies not satisfied, returns a new NLU pipeline where dependencies and sub-dependencies are satisfied.
        Checks and resolves in the following order :
        1. Get a reference list of input features missing for the current pipe
        2. Resolve the list of missing features by adding new  Annotators to pipe
        3. Add NER Converter if required (When there is a NER model)
        4. Fix order and output column names
        5.

        :param pipe:
        :return:
        """
        # main entry point for Model stacking withouth pretrained pipelines
        # requirements and provided features will be lists of lists

        #0. Clean old @AT storage ref from all columns
        logger.info('Cleaning old AT refs')
        pipe = PipeUtils.clean_AT_storage_refs(pipe)

        #1. Resolve dependencies, builds a DAG in reverse and satisfies dependencies with a Breadth-First-Search approach
        logger.info('Satisfying dependencies')
        pipe = PipelineQueryVerifier.satisfy_dependencies(pipe)

        #2. Enforce naming schema <col_name>@<storage_ref> for storage_ref consumers and producers and <entity@nlu_ref> and <ner@nlu_ref> for NER and NER-Converters
        # and add NER-IOB to NER-Pretty converters for every NER model that is not already feeding a NER converter
        pipe = PipeUtils.enforce_AT_schema_on_pipeline(pipe)

        #3. Validate naming of output columns is correct and no error will be thrown in spark
        logger.info('Fixing column names')
        pipe = PipelineQueryVerifier.check_and_fix_component_output_column_name_satisfaction(pipe)

        #4. Set on every NLP Annotator the output columns
        pipe = PipeUtils.enforce_NLU_columns_to_NLP_columns(pipe)

        #5. fix order
        logger.info('Optimizing pipe component order')
        pipe = PipelineQueryVerifier.check_and_fix_component_order(pipe)

        #6. Rename overlapping/duplicate leaf columns in the DAG
        logger.info('Renaming duplicates cols')
        pipe = PipeUtils.rename_duplicate_cols(pipe)

        #7. enfore again because trainable pipes might mutate pipe cols
        pipe = PipeUtils.enforce_NLU_columns_to_NLP_columns(pipe)

        logger.info('Done with pipe optimizing')

        return pipe




    @staticmethod
    def check_and_fix_component_order(pipe):
        '''
        This method takes care that the order of components is the correct in such a way,that the pipeline can be iteratively processed by spark NLP.
        Column Names will not be touched. DAG Task Sort basically.
        '''
        logger.info("Starting to optimize component order ")
        correct_order_component_pipeline = []
        all_components_orderd = False
        all_components = pipe.components
        provided_features = []
        update_last_type = False
        last_type_sorted = None
        trainable_updated = False
        while all_components_orderd == False:
            if update_last_type : last_type_sorted = None
            else : update_last_type = True
            for component in all_components:
                logger.info(f"Optimizing order for component {component.info.name}")
                # input_columns = ComponentUtils.clean_irrelevant_features(component.info.spark_input_column_names, False)
                input_columns = ComponentUtils.remove_storage_ref_from_features(
                    ComponentUtils.clean_irrelevant_features(component.info.spark_input_column_names,False))


                if last_type_sorted is None or component.info.type == last_type_sorted:
                    if set(input_columns).issubset(provided_features):
                        correct_order_component_pipeline.append(component)
                        if component in all_components: all_components.remove(component)
                        # for feature in component.info.spark_output_column_names: provided_features.append(feature)

                        # provided_features += ComponentUtils.clean_irrelevant_features(component.info.spark_output_column_names,False)
                        # TODO REMOVE STORAGE REF FROM PROVDED FEATURES???
                        provided_features += ComponentUtils.remove_storage_ref_from_features(
                            ComponentUtils.clean_irrelevant_features(component.info.spark_output_column_names,False))

                        last_type_sorted = component.info.type
                        update_last_type = False
                        break
            if len(all_components) == 0: all_components_orderd = True

            if not all_components_orderd  and len(all_components) <= 2 and pipe.has_trainable_components and not trainable_updated  and 'approach' in str(all_components[0].model).lower() and 'sentence_embeddings@' in all_components[0].info.inputs:
                # special case, if trainable then we feed embed consumers on the first sentence embed provider
                # 1. Find first sent embed provider
                # 2. substitute any 'sent_embed@' consumer inputs for the provider col
                for f in provided_features:
                    if 'sentence_embeddings' in f and not trainable_updated  :
                        all_components[0].info.spark_input_column_names.remove('sentence_embeddings@')
                        if 'sentence_embeddings@' in  all_components[0].info.inputs :  all_components[0].info.inputs.remove('sentence_embeddings@')
                        all_components[0].info.spark_input_column_names.append(f)
                        if f not in all_components[0].info.inputs :  all_components[0].info.inputs.append(f)
                        trainable_updated = True

            if not all_components_orderd and len(all_components) <= 2 and pipe.has_trainable_components and not trainable_updated  and 'approach' in str(all_components[0].model).lower() and 'word_embeddings@' in all_components[0].info.inputs:
                # special case, if trainable then we feed embed consumers on the first sentence embed provider
                # 1. Find first sent embed provider
                # 2. substitute any 'sent_embed@' consumer inputs for the provider col
                for f in provided_features:
                    if 'word_embeddings' in f and not trainable_updated  :
                        all_components[0].info.spark_input_column_names.remove('word_embeddings@')
                        if 'word_embeddings@' in  all_components[0].info.inputs :  all_components[0].info.inputs.remove('word_embeddings@')
                        all_components[0].info.spark_input_column_names.append(f)
                        if f not in all_components[0].info.inputs :  all_components[0].info.inputs.append(f)
                        trainable_updated = True


        pipe.components = correct_order_component_pipeline

        return pipe




    @staticmethod
    def log_resolution_status(provided_features_no_ref,required_features_no_ref,provided_features_ref,required_features_ref,is_trainable,conversion_candidates,missing_features_no_ref,missing_features_ref,):
        logger.info(f"========================================================================")
        logger.info(f"Resolution Status provided_features_no_ref = {set(provided_features_no_ref)}")
        logger.info(f"Resolution Status required_features_no_ref = {set(required_features_no_ref)}")
        logger.info(f"Resolution Status provided_features_ref    = {set(provided_features_ref)}")
        logger.info(f"Resolution Status required_features_ref    = {set(required_features_ref)}")
        logger.info(f"Resolution Status is_trainable             = {is_trainable}")
        logger.info(f"Resolution Status conversion_candidates    = {conversion_candidates}")
        logger.info(f"Resolution Status missing_features_no_ref  = {set(missing_features_no_ref)}")
        logger.info(f"Resolution Status conversion_candidates    = {set(missing_features_ref)}")
        logger.info(f"========================================================================")



    @staticmethod
    def is_storage_ref_match(embedding_consumer, embedding_provider,pipe):
        """Check for 2 components, if one provides the embeddings for the other. Makes sure that output_level matches up (chunk/sent/tok/embeds)"""
        consumer_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(embedding_consumer, 'input')
        provider_AT_rev = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(embedding_provider, 'output')
        consum_level    = ComponentUtils.extract_embed_level_identity(embedding_consumer, 'input')
        provide_level   = ComponentUtils.extract_embed_level_identity(embedding_provider, 'output')

        consumer_ref    = StorageRefUtils.extract_storage_ref(embedding_consumer)
        provider_rev    = StorageRefUtils.extract_storage_ref(embedding_provider)

        # input/output levels must match
        if consum_level != provide_level : return False

        # If storage ref dont match up, we must consult the storage_ref_2_embed mapping if it still maybe is a match, otherwise it is not.
        if consumer_ref == provider_rev  : return True

        # Embed Components have have been resolved via@ have a  nlu_resolution_ref_source will match up with the consumer ref if correct embedding.
        if hasattr(embedding_provider.info, 'nlu_ref'):
            if consumer_ref == StorageRefUtils.extract_storage_ref(embedding_provider.info.nlu_ref) : return True

        # If it is either  sentence_embedding_converter or chunk_embedding_converter then we gotta check what the storage ref of the inpot of those is.
        # If storage ref matches up, the providers output will match the consumer
        # if embedding_provider
        if embedding_provider.info.name in ["chunk_embedding_converter", 'sentence_embedding_converter']: # TODO FOR RESOLUTION
            nlu_ref, conv_prov_storage_ref = PipelineQueryVerifier.get_converters_provider_info(embedding_provider,pipe)


        return False





    @staticmethod
    def is_matching_level(embedding_consumer, embedding_provider):
        """Check for embedding consumer if input level matches up outputlevel of consumer
        """

    @staticmethod
    def get_converters_provider_info(embedding_provider,pipe):
        """For a component and a pipe, find storage_ref and """




"""
Test Any open source annotator class with this module.
you just need to call verify_uploaded_model_is_is_correct()
Param Info for verify_uploaded_model_is_is_correct():
ONLY USE either nlu_referenced_requirements OR nlp_referenced_requirements please.
model_path : str
    Path to where the component to load is. Either path to Pipeline or Model
nlu_referenced_requirements : List[str]
    a List of requirements. Should either be NLU or Spark NLP references to specific models, i.e. BIOBERT embeddings
nlp_referenced_requirements : List[Tuple[str,str]]
    A list of Tuples, where first element is SparkNLP Model reference and second Element is Language.



See
test_load_and_verify_pipeline()
test_load_and_verify_model_with_requirements_nlu_based()
test_load_and_verify_model_with_requirements_spark_based()
test_load_and_verify_model_with_no_storage_ref_requirement()
for example usage
"""
import nlu.pipe.component_resolution

""" Map SparkNLP Annotator Classes to load methods which will return corrosponding models when called with parameter
This map is actually not used because of the eval() call!
"""
# from sparknlp.annotator import *
# d = {
#     MultiClassifierDLModel : MultiClassifierDLModel.load ,
#     PerceptronModel : PerceptronModel.load,
#     ClassifierDLModel : ClassifierDLModel.load,
#     BertEmbeddings : BertEmbeddings.load,
#     AlbertEmbeddings : AlbertEmbeddings.load,
#     XlnetEmbeddings : XlnetEmbeddings.load,
#     WordEmbeddingsModel : WordEmbeddingsModel.load,
#     ElmoEmbeddings : ElmoEmbeddings.load,
#     BertSentenceEmbeddings : BertSentenceEmbeddings.load,
#     UniversalSentenceEncoder : UniversalSentenceEncoder.load,
#     TokenizerModel : TokenizerModel.load,
#     DocumentAssembler : DocumentAssembler.load,
#     SentenceDetectorDLModel : SentenceDetectorDLModel.load,
#     SentenceDetector : SentenceDetector.load,
#     ContextSpellCheckerModel : ContextSpellCheckerModel.load,
#     SymmetricDeleteModel : SymmetricDeleteModel.load,
#     NorvigSweetingModel : NorvigSweetingModel.load,
#     LemmatizerModel : LemmatizerModel.load,
#     NormalizerModel : NormalizerModel.load,
#     Stemmer : Stemmer.load,
#     NerDLModel : NerDLModel.load,
#     NerCrfModel : NerCrfModel.load,
#     LanguageDetectorDL : LanguageDetectorDL.load,
#     DependencyParserModel : DependencyParserModel.load,
#     TypedDependencyParserModel : TypedDependencyParserModel.load,
#     SentimentDLModel : SentimentDLModel.load,
#     SentimentDetectorModel : SentimentDetectorModel.load,
#     ViveknSentimentModel : ViveknSentimentModel.load,
#     Chunker : Chunker.load,
#     NGramGenerator : NGramGenerator.load,
#     ChunkEmbeddings : ChunkEmbeddings.load,
#     StopWordsCleaner : StopWordsCleaner.load,
#     TextMatcherModel : TextMatcherModel.load,
#     RegexMatcherModel : RegexMatcherModel.load,
#     DateMatcher : DateMatcher.load,
#     MultiDateMatcher : MultiDateMatcher.load,
#     T5Transformer : T5Transformer.load,
#     MarianTransformer : MarianTransformer.load,
#     PretrainedPipeline : PipelineModel.load,

## NEED MODEL VERSION OF THESE saved ( LICENSED!)
# ContextualParserModel
# - AssertionDLApproach
# - ChunkEntityResolverApproach
# - SentenceEntityResolverApproach
# - DeIdentification
# - RelationExtractionApproach  (NO DL VERSION!!)
# - GenericClassifierApproach  x?
# }

import json
from nlu.pipe.pipe_logic import PipelineQueryVerifier
from nlu.pipe.pipeline import  *
import os
from nlu.pipe.pipe_components import SparkNLUComponent
from pyspark.ml import PipelineModel
def NLP_ref_to_NLU_ref(nlp_ref,lang) :
    """Resolve a Spark NLP reference to a NLU reference"""
    nlu_namespaces_to_check = [nlu.Spellbook.pretrained_pipe_references, nlu.Spellbook.pretrained_models_references]
    for dict_ in nlu_namespaces_to_check:
        if lang in dict_.keys():
            for reference in dict_[lang]:
                if dict_[lang][reference] == nlp_ref:
                    return reference

def check_is_pipeline(path):
    # Folder should contain Stages
    if 'stages' in os.listdir(path) : return True
    return False
def check_is_model(path):
    # Folder should contain Metadata
    if 'metadata' in os.listdir(path) : return True
    return False
def verify_uploaded_model_is_is_correct(model_path:str,nlu_referenced_requirements = [], nlp_referenced_requirements = []):
    """ Use a Annotator Class and a list of either NLU or NLP requirement references to automatically build a Pipeline.
    ONLY USE either nlu_referenced_requirements OR nlp_referenced_requirements please.
    model_path : str
        Path to where the component to load is. Either path to Pipeline or Model
    nlu_referenced_requirements : List[str]
        a List of requirements. Should either be NLU or Spark NLP references to specific models, i.e. BIOBERT embeddings
    nlp_referenced_requirements : List[Tuple[str,str]]
        A list of Tuples, where first element is SparkNLP Model reference and second Element is Language.


    # 1. Build NLU_component class, CANNOT use nrmal builders, need special construction. REQUIRE INFOS FOR build_and_fix_pie process!
    # 2. Build dependency classes i.e. embeddings NLU_components via nlu.load() sub_emthod and get the resolved comonent
    # 3.
    # 2. Build Pipeline with trained anno + requiredments and run "Fix_urself()" magic call
    # 3. Build a Spark DF and RUn pipe, GG
    # 4. pythonfiy>?
    """
    if check_is_pipeline(model_path):
        return verify_pipeline(model_path)
    elif check_is_model(model_path):
            return verify_model(model_path,nlu_referenced_requirements, nlp_referenced_requirements)
    else:
        print("Neither STAGES or METADATA folder detected in unzipped folder. Make sure you include these folders")
        return False

def verify_pipeline(model_path:str,test_data = 'Hello world from John Snow Labs'):
    """
     Build pyspark Pipeline and transform data. No NLU usage
    """
    spark = sparknlp.start()
    df = spark.createDataFrame(pd.DataFrame({'text': test_data}, index=[0]))
    pipe = PipelineModel.load(model_path)
    res = pipe.transform(df)
    res.show()
    return res
def verify_model(model_path:str,nlu_referenced_requirements = [], nlp_referenced_requirements = [],test_data='Hello world from John Snow Labs '):
    """
     Build model with requirements
     Figures out class name by checking metadata json file
     assumes metadatra is always called part-00000
    """
    with open(model_path+'/metadata/'+'part-00000') as json_f:
        class_name = json.load(json_f)['class'].split('.')[-1]
        # The last element in the Class name can be used to just load the model from disk!
        # Just call eval on it, which will give you the actual Python class reference which should have a .load() method
        m = eval(class_name).load(model_path)
    component_type,nlu_anno_class, = resolve_annotator_class_to_nlu_component_info(class_name)
    # Wrap model with NLU Custom Model class so the NLU pipeline Logic knows what to do with it
    c = CustomModel(annotator_class=nlu_anno_class, component_type=component_type, model=m)
    pipe = NLUPipeline()
    pipe.add(c)

    # get requirements
    if len(nlu_referenced_requirements)==0 and len(nlp_referenced_requirements) == 0 :
        pipe = PipelineQueryVerifier.check_and_fix_nlu_pipeline(pipe)
        res = pipe.predict(test_data)
        print(res)
        return res
    elif len(nlu_referenced_requirements) >0:
        for r in nlu_referenced_requirements:
            pipe.add(nlu.pipe.component_resolution.nlu_ref_to_component(r))
    elif len(nlp_referenced_requirements) > 0:
        for r in nlu_referenced_requirements:
            # map back to NLU ref
            pipe.add(nlu.pipe.component_resolution.nlu_ref_to_component(NLP_ref_to_NLU_ref(r)))
    # run pipe with dependencies
    pipe = PipelineQueryVerifier.check_and_fix_nlu_pipeline(pipe)
    res = pipe.predict(test_data)
    print(res)
    return res
def resolve_annotator_class_to_nlu_component_info(anno_class ='LemmatizerModel'):
    """
    SparkNLUComponent.__init__(self, annotator_class, component_type)
    RECURISIVELY SEARCH through NLU COMPONENTS FOLDER for each class for a given <CLASSS>
    Find the file, which called <CLASS>.pretrained or just <CLASS> !!!
    In the folder containing that found file there will be the component_json info we need!]\
    """
    p = nlu.nlu_package_location+'components/'
    anno_class = anno_class+'.'

    import os
    for dirpath, dirs, files in os.walk(p):
        # search for folder that has the component info for that anno_class
        for filename in files:
            if '.py' not in filename or '.pyc' in filename : continue
            fname = os.path.join(dirpath,filename)
            if test_check_if_string_in_file(fname, anno_class):
                parts = dirpath.split('/')
                component_type = parts[-2]
                nlu_anno_class = parts[-1]
                component_type = component_type[:-1]
                return component_type,nlu_anno_class,
    print("COULD NOT FIND COMPONENT INFO FOR ANNO_CLASS", anno_class)
    return False

def test_check_if_string_in_file(file_name, string_to_search):
    """ Check if any line in the file contains given string """
    # Open the file in read only mode
    # print('reading ', file_name)
    with open(file_name, 'r') as read_obj:
        # Read all lines in the file one by one
        for line in read_obj:
            # For each line, check if line contains the string
            if string_to_search in line:
                return True
    return False

class CustomModel(SparkNLUComponent):
    """ Builds a NLU Components with component info"""
    def __init__(self, annotator_class='sentiment_dl',  component_type='classifier', model = None):
        self.model = model
        SparkNLUComponent.__init__(self, annotator_class, component_type)
        # Make sure input/output cols match up with NLU defaukts
        if len(self.info.spark_input_column_names) == 1 :
            model.setInputCols(self.info.spark_input_column_names[0])
        else :
            model.setInputCols(self.info.spark_input_column_names)

        if len(self.info.spark_output_column_names) == 1 :
            model.setOutputCol(self.info.spark_output_column_names[0])
        else :
            model.setOutputCol(self.info.spark_output_column_names)

def test_load_and_verify_pipeline(): # OK
    # nlu.load('ner.onto') # Load a pipe to ~/cache_pretrained folder
    pipe_path='/home/loan/cache_pretrained/onto_recognize_entities_sm_en_2.4.0_2.4_1579730599257'
    res = verify_uploaded_model_is_is_correct(pipe_path)
    print(res.columns)
    res.show()
    if res is None :  return False # MODEL CAUSE A CRASH!
    else :  return True

def test_load_and_verify_model_with_requirements_nlu_based(): # OK
    # nlu.load('en.sentiment.twitter.use') # Load a model with storage ref requriements to your ~/cache_pretraiend folder
    model_path = '/home/loan/cache_pretrained/analyze_sentimentdl_use_twitter_en_2.7.1_2.4_1610993470852/stages/2_SentimentDLModel_eca587b575f7' #   Sentiment DL with USE requirements
    # requirements = ['use']
    # NLU can auto-resolve some storage_refs, but it will not work every time! Better be safe and pass actual reference to requirements
    res = verify_uploaded_model_is_is_correct(model_path)#,requirements)
    print(res.columns)
    print(res)
    if res is None :  return False # MODEL CAUSED A CRASH!
    else :  return True

def test_load_and_verify_model_with_no_storage_ref_requirement(): # OK
    # nlu.load('lemma') # Load a model with no storage ref to ~/cache_pretrained folder
    model_path = '/home/loan/cache_pretrained/lemma_bh_2.7.0_2.4_1610989221391' # Sample lemmatizer
    res = verify_uploaded_model_is_is_correct(model_path)
    print(res.columns)
    print(res)
    if res is None :  return False # MODEL CAUSED A CRASH!
    else :  return True

def test_load_and_verify_model_with_requirements_spark_based(): # OK
    # nlu.load('en.sentiment.twitter.use') # Load a model with storage ref requriements to your ~/cache_pretraiend folder
    model_path = '/home/loan/cache_pretrained/analyze_sentimentdl_use_twitter_en_2.7.1_2.4_1610993470852/stages/2_SentimentDLModel_eca587b575f7' #   Sentiment DL with USE requirements
    requirements = ['tfhub_use','en']
    # NLU can auto-resolve some storage_refs, but it will not work every time! Better be safe and pass actual reference to requirements
    res = verify_uploaded_model_is_is_correct(model_path,nlp_referenced_requirements=requirements)#,requirements)
    print(res.columns)
    print(res)
    if res is None :  return False # MODEL CAUSED A CRASH!
    else :  return True

# Test every verification method. Only Open Source annotators and dependencies will work.
test_load_and_verify_model_with_requirements_spark_based()

test_load_and_verify_pipeline()
test_load_and_verify_model_with_requirements_nlu_based()
test_load_and_verify_model_with_no_storage_ref_requirement()










































import unittest
from nlu import *

class TestPOS(unittest.TestCase):

    def test_pos_model(self):
        df = nlu.load('pos',verbose=True).predict('Women belong in the kitchen')
        for c in df.columns: print(df[c])




if __name__ == '__main__':
    unittest.main()




from sparknlp_jsl.annotator import *
def verify_model_licensed(class_name : str, model_path:str):
    """
    Load a licensed model from HDD
    """
    try :
        m = eval(class_name).load(model_path)
        return m
    except:
        print(f"Could not load Annotator class={class_name} located in {model_path}. Try updaing spark-nlp-jsl")



from sparknlp.annotator import NerConverter,DependencyParserModel
from typing import List, Tuple, Optional, Dict
import streamlit as st
from nlu.utils.modelhub.modelhub_utils import ModelHubUtils
import numpy as np
import pandas as pd
from sparknlp.annotator import *
import nlu
class StreamlitUtilsOS():
    classifers_OS = [ ClassifierDLModel, LanguageDetectorDL, MultiClassifierDLModel, NerDLModel, NerCrfModel, YakeModel, PerceptronModel, SentimentDLModel,
                      SentimentDetectorModel, ViveknSentimentModel, DependencyParserModel, TypedDependencyParserModel, T5Transformer, MarianTransformer, NerConverter]
    @staticmethod
    def get_classifier_cols(pipe):
        classifier_cols = []
        for c in pipe.components:
            if type(c.model) in StreamlitUtilsOS.classifers_OS :
                classifier_cols += pipe.anno2final_cols[c.model]
        return classifier_cols

    @staticmethod
    def get_embed_cols(pipe):
        classifier_cols = []
        embedders = StreamlitUtilsOS.find_all_embed_components(pipe)
        for c in embedders: classifier_cols += pipe.anno2final_cols[c.model]
        return classifier_cols

    @staticmethod
    def find_embed_col(df, search_multi=False):
        """Find col that contains embed"""
        if not search_multi:
            for c in df.columns:
                if 'embed'in c : return c
        else:
            e_cols =[]
            print("DDDDDDDDDDDDDDDDDDDDDDDEBUG ::". df.columns)
            for c in df.columns:
                if 'embed'in c : e_cols.append(c)
        return e_cols


    @staticmethod
    def find_embed_component(p):
        """Find first embed  component in pipe"""
        for c in p.components :
            if 'embed' in c.info.outputs[0] : return c
        st.warning("No Embed model in pipe")
        return None

    @staticmethod
    def find_all_classifier_components(pipe):
        """Find ALL classifier component in pipe"""
        classifier_comps = []
        for c in pipe.components:
            if type(c.model) in StreamlitUtilsOS.classifers_OS :classifier_comps.append(c)
        return  classifier_comps
    @staticmethod
    def find_all_embed_components(p):
        """Find ALL  embed component in pipe"""
        cs = []
        for c in p.components :
            if 'embed' in c.info.outputs[0] and 'chunk' not  in c.info.outputs[0]: cs.append(c)
        if len(cs) == 0 : st.warning("No Embed model in pipe")
        return cs

    @staticmethod
    def extract_name(component_or_pipe):
        name =''
        if hasattr(component_or_pipe,'info') :
            if hasattr(component_or_pipe.info,'nlu_ref') : name = component_or_pipe.info.nlu_ref
            elif hasattr(component_or_pipe,'storage_ref') : name = component_or_pipe.info.storage_ref
            elif hasattr(component_or_pipe,'nlp_ref') : name = component_or_pipe.info.nlp_ref
        elif hasattr(component_or_pipe,'nlu_ref') : name = component_or_pipe.nlu_ref
        return name


    @staticmethod
    def find_ner_model(p):
        """Find NER component in pipe"""
        from sparknlp.annotator import NerDLModel,NerCrfModel
        for c in p.components :
            if isinstance(c.model,(NerDLModel,NerCrfModel)):return c.model
        st.warning("No NER model in pipe")
        return None

    @staticmethod
    def get_NER_tags_in_pipe(p):
        """Get NER tags in pipe, used for showing visualizable tags"""
        n = StreamlitUtilsOS.find_ner_model(p)
        if n is None : return []
        classes_predicted_by_ner_model = n.getClasses()
        split_iob_tags = lambda s : s.split('-')[1] if '-' in s else ''
        classes_predicted_by_ner_model = list(map(split_iob_tags,classes_predicted_by_ner_model))
        while '' in classes_predicted_by_ner_model : classes_predicted_by_ner_model.remove('')
        classes_predicted_by_ner_model = list(set(classes_predicted_by_ner_model))
        return classes_predicted_by_ner_model

    @staticmethod
    def get_manifold_algo(algo,dim, n_jobs = None):
        from sklearn.manifold import TSNE, Isomap, LocallyLinearEmbedding, MDS, SpectralEmbedding
        from sklearn.decomposition import TruncatedSVD,DictionaryLearning, FactorAnalysis, FastICA, KernelPCA, PCA
        # manifold
        if algo=='TSNE' : return TSNE(n_components=dim, n_jobs=n_jobs)
        if algo=='ISOMAP' : return Isomap(n_components=dim, n_jobs=n_jobs)
        if algo=='LLE' : return LocallyLinearEmbedding(n_components=dim, n_jobs=n_jobs)
        if algo=='Spectral Embedding' : return  SpectralEmbedding(n_components=dim, n_jobs=n_jobs)
        if algo=='MDS' : return MDS(n_components=dim, n_jobs=n_jobs)
        # Matrix Decomposition
        if algo== 'PCA' : return PCA(n_components=dim) # No hyper
        if algo== 'SVD aka LSA' : return TruncatedSVD(n_components=dim) # No hyper
        if algo =='DictionaryLearning': return DictionaryLearning(n_components=dim, n_jobs=n_jobs)
        if algo =='FactorAnalysis': return FactorAnalysis(n_components=dim) # no hyper
        if algo =='FastICA': return FastICA(n_components=dim) # no hyper
        if algo =='KernelPCA': return KernelPCA(n_components=dim, n_jobs=n_jobs)
        # not applicable because negative values, todo we could just take absolute values of all embeds..
        # if algo =='LatentDirichletAllocation': return LatentDirichletAllocation(n_components=dim)
        # if algo =='NMF': return NMF(n_components=dim)



    @staticmethod
    @st.cache(allow_output_mutation=True)
    def get_pipe(model='ner'): return nlu.load(model)

    @staticmethod
    def merge_token_classifiers_with_embed_pipe(embed_pipe, token_pipe):
        """Merge token feature generators into embed pipe. i.e. Pos/Dep_depdency/Untyped_dep if not already present in pipe"""
        for c in token_pipe.components :
            if c.info.name == 'pos' :
                for emb_c in embed_pipe.components :
                    if emb_c.info.name =='pos' : return embed_pipe
                    # only merge if pos not already in pipe
                embed_pipe.components.append(c)
            # if c.info.name == 'unlabeled_dependency_parser' : embed_pipe.components.append(c)
            # if c.info.name == 'labeled_dependency_parser' : embed_pipe.components.append(c)
        return embed_pipe




import unittest
from tests.test_utils import get_sample_pdf_with_labels, get_sample_pdf, get_sample_sdf, get_sample_pdf_with_extra_cols, get_sample_pdf_with_no_text_col ,get_sample_spark_dataframe
from nlu import *

class TestYake(unittest.TestCase):

    def test_yake_model(self):
        #setting meta to true will output scores for keywords. Lower scores are better
        df = nlu.load('yake',verbose=True).predict('What a wonderful day! Arnold schwanenegger is the Terminator and he wants to get to the American chopper', output_level='sentence',drop_irrelevant_cols=False, metadata=True, )
        for c in df.columns: print(df[c])
        df = nlu.load('yake',verbose=True).predict('What a wonderful day! Arnold schwanenegger is the Terminator and he wants to get to the American chopper', metadata=False)
        for c in df.columns: print(df[c])
        df = nlu.load('yake',verbose=True).predict('What a wonderful day! Arnold schwanenegger is the Terminator and he wants to get to the American chopper', output_level='token')
        for c in df.columns: print(df[c])
        df = nlu.load('yake',verbose=True).predict('What a wonderful day! Arnold schwanenegger is the Terminator and he wants to get to the American chopper', output_level='chunk')
        for c in df.columns: print(df[c])
        #Column name of confidence changed if yake at same or not at same output level!
        df = nlu.load('yake',verbose=True).predict('What a wonderful day! Arnold schwanenegger is the Terminator and he wants to get to the American chopper', output_level='document')
        for c in df.columns: print(df[c])

if __name__ == '__main__':
    unittest.main()




import streamlit as st
import pandas as pd
import numpy as np



"""@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@MAP AND BART CHART BELOW @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"""

st.title('Uber pickups in NYC')

DATE_COLUMN = 'date/time'
DATA_URL = ('https://s3-us-west-2.amazonaws.com/'
            'streamlit-demo-data/uber-raw-data-sep14.csv.gz')

@st.cache
def load_data(nrows):
    data = pd.read_csv(DATA_URL, nrows=nrows)
    lowercase = lambda x: str(x).lower()
    data.rename(lowercase, axis='columns', inplace=True)
    data[DATE_COLUMN] = pd.to_datetime(data[DATE_COLUMN])
    return data

data_load_state = st.text('Loading data...')
data = load_data(10000)
data_load_state.text("Done! (using st.cache)")

if st.checkbox('Show raw data'):
    st.subheader('Raw data')
    st.write(data)

st.subheader('Number of pickups by hour')
hist_values = np.histogram(data[DATE_COLUMN].dt.hour, bins=24, range=(0,24))[0]
st.bar_chart(hist_values)

# Some number in the range 0-23
hour_to_filter = st.slider('hour', 0, 23, 17)
filtered_data = data[data[DATE_COLUMN].dt.hour == hour_to_filter]

st.subheader('Map of all pickups at %s:00' % hour_to_filter)
st.map(filtered_data)
"""@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ALTAIOR REAL TIME PLOT BELOW@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"""
import altair as alt
import numpy as np
import pandas as pd
import streamlit as st
import time
# Generate some random data
df = pd.DataFrame(np.random.randn(20, 3), columns=["a", "b", "c"])

# Build a scatter chart using altair. I modified the example at
# https://altair-viz.github.io/gallery/scatter_tooltips.html
scatter_chart = st.altair_chart(
    alt.Chart(df)
        .mark_circle(size=60)
        .encode(x='a', y='b', color='c')
        .interactive()
)

# Append more random data to the chart using add_rows
for ii in range(0, 100):
    df = pd.DataFrame(np.random.randn(20, 3), columns=["a", "b", "c"])
    scatter_chart.add_rows(df)
    # Sleep for a moment just for demonstration purposes, so that the new data
    # animates in.
    time.sleep(0.1)

"""@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@3d  THINLGY BELOW https://discuss.streamlit.io/t/plotly-in-streamlit/1319/4 @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@"""

import streamlit as st
import numpy as np
import pandas as pd
import plotly.graph_objects as go

# from Inputs_Parallel import get_possible_scenarios

# Side Bar #######################################################
project_title = st.sidebar.text_input(label="Title of Project",
                                      value="Example Project")

username = st.sidebar.selectbox(label="Username",
                                options=("a_name",
                                         "b_name"))

buildable_land_folder = st.sidebar.text_input(label="Buildable Land Folder",
                                              value=r"\\filepath\example")

config_file_location = st.sidebar.text_input(label="Config File",
                                             value=r"\\filepath\example")

gcr_config = st.sidebar.slider(label="Ground Coverage Ratio Range Selection",
                               min_value=10,
                               max_value=60,
                               step=1,
                               value=(28, 45))

sr_config = st.sidebar.slider(label="Sizing Ratio Range Selection",
                              min_value=1.0,
                              max_value=2.0,
                              step=0.1,
                              value=(1.0, 1.5))

run_button = st.sidebar.button(label='Run Optimization')

progress_bar = st.sidebar.progress(0)

# App ###########################################################
st.title(project_title)

# Graphing Function #####
z_data = pd.read_csv('https://raw.githubusercontent.com/plotly/datasets/master/api_docs/mt_bruno_elevation.csv')
z = z_data.values
sh_0, sh_1 = z.shape
x, y = np.linspace(0, 1, sh_0), np.linspace(0, 1, sh_1)
fig = go.Figure(data=[go.Surface(z=z, x=x, y=y)])
fig.update_layout(title='IRR', autosize=False,
                  width=800, height=800,
                  margin=dict(l=40, r=40, b=40, t=40))
st.plotly_chart(fig)


from sparknlp.annotator import *
import inspect
import logging

logger = logging.getLogger('nlu')
from nlu.pipe.utils.component_utils import ComponentUtils
class PipeUtils():
    """Pipe Level logic oprations and utils"""
    @staticmethod
    def is_trainable_pipe(pipe):
        '''Check if pipe is trainable'''
        for c in pipe.components:
            if ComponentUtils.is_untrained_model(c):return True
        return False



    @staticmethod
    def enforece_AT_embedding_provider_output_col_name_schema_for_list_of_components  (pipe_list):
        """For every embedding provider, enforce that their output col is named <output_level>@storage_ref for output_levels word,chunk,sentence aka document , i.e. word_embed@elmo or sentence_embed@elmo etc.."""
        for c in pipe_list:
            if ComponentUtils.is_embedding_provider(c):
                level_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(c, 'output')
                c.info.outputs = [level_AT_ref]
                c.info.spark_output_column_names = [level_AT_ref]
                c.model.setOutputCol(level_AT_ref[0])
        return pipe_list


    @staticmethod
    def enforce_AT_schema_on_pipeline(pipe):
        """Enforces the AT naming schema on all column names and add missing NER converters"""
        return PipeUtils.enforce_AT_schema_on_NER_processors_and_add_missing_NER_converters(PipeUtils.enforce_AT_schema_on_embedding_processors(pipe))

    @staticmethod
    def enforce_AT_schema_on_NER_processors_and_add_missing_NER_converters(pipe):
        """For every NER provider and consumer, enforce that their output col is named <output_level>@storage_ref for output_levels word,chunk,sentence aka document , i.e. word_embed@elmo or sentence_embed@elmo etc..
        We also add NER converters for every NER model that no Converter converting it's inputs
        In addition, returns the pipeline with missing NER converters added, for every NER model.
        The converters transform the IOB schema in a merged and more usable form for downstream tasks
        1. Find a NER model in pipe
        2. Find a NER converter feeding from it, if there is None, create one.
        3. Generate name with Identifier  <ner-iob>@<nlu_ref_identifier>  and <entities>@<nlu_ref_identifier>
        3.1 Update NER Models    output to <ner-iob>@<nlu_ref_identifier>
        3.2 Update NER Converter input  to <ner-iob>@<nlu_ref_identifier>
        3.3 Update NER Converter output to <entities>@<nlu_ref_identifier>
        4. Update every Component that feeds from the NER converter (i.e. Resolver etc..)
        """
        from nlu import Util
        new_converters = []
        for c in pipe.components:
            if ComponentUtils.is_NER_provider(c):
                output_NER_col = ComponentUtils.extract_NER_col(c,'output')
                converter_to_update = None
                # if '@' not in output_NER_col:
                for other_c in pipe.components:
                    if output_NER_col in other_c.info.inputs and ComponentUtils.is_NER_converter(other_c):
                        converter_to_update = other_c

                ner_identifier = ComponentUtils.get_nlu_ref_identifier(c)
                if converter_to_update is  None :
                    if c.info.license == 'healthcare': converter_to_update  = Util("ner_to_chunk_converter_licensed",is_licensed=True )
                    else : converter_to_update  = Util("ner_to_chunk_converter")
                    new_converters.append(converter_to_update)

                # converter_to_update.info.nlu_ref = f'ner_converter@{ner_identifier}'
                converter_to_update.info.nlu_ref = f'ner_converter.{c.info.nlu_ref}'

                # 3. generate new col names
                new_NER_AT_ref = output_NER_col
                if '@' not in output_NER_col: new_NER_AT_ref = output_NER_col + '@' + ner_identifier
                new_NER_converter_AT_ref = 'entities' + '@' + ner_identifier

                # 3.1 upate NER model outputs
                c.info.outputs = [new_NER_AT_ref]
                c.info.spark_output_column_names = [new_NER_AT_ref]
                c.model.setOutputCol(new_NER_AT_ref)

                #3.2 update converter inputs
                old_ner_input_col = ComponentUtils.extract_NER_converter_col(converter_to_update, 'input')
                if old_ner_input_col in converter_to_update.info.inputs:
                    converter_to_update.info.inputs.remove(old_ner_input_col)
                else : converter_to_update.info.inputs.pop()

                # converter_to_update.info.inputs.remove(old_ner_input_col)
                if old_ner_input_col in converter_to_update.info.spark_input_column_names :
                    converter_to_update.info.spark_input_column_names.remove(old_ner_input_col)
                else : converter_to_update.info.spark_input_column_names.pop()
                converter_to_update.info.inputs.append(new_NER_AT_ref)
                converter_to_update.info.spark_input_column_names.append(new_NER_AT_ref)
                converter_to_update.model.setInputCols(converter_to_update.info.inputs)

                #3.3 update converter outputs
                converter_to_update.info.outputs = [new_NER_converter_AT_ref]
                converter_to_update.info.spark_output_column_names = [new_NER_converter_AT_ref]
                converter_to_update.model.setOutputCol(new_NER_converter_AT_ref)

                ## todo improve, this causes the first ner producer to feed to all ner-cosnuners. All other ner-producers will be ignored by ner-consumers,w ithouth special syntax or manual configs --> Chunk merger
                ##4. Update all NER consumers input columns, i.e. Resolver, Relation, etc..
                for conversion_consumer in pipe.components :
                    if 'entities' in conversion_consumer.info.inputs:
                        conversion_consumer.info.inputs.remove('entities')
                        conversion_consumer.info.spark_input_column_names.remove('entities')
                        conversion_consumer.info.inputs.append(new_NER_converter_AT_ref)
                        conversion_consumer.info.spark_input_column_names.append(new_NER_converter_AT_ref)

        # Add new converters to pipe
        for conv in new_converters:
            if conv.info.license == 'healthcare':
                pipe.add(conv,  name_to_add = f'chunk_converter_licensed@{conv.info.outputs[0].split("@")[0]}')
            else :
                pipe.add(conv,  name_to_add = f'chunk_converter@{conv.info.outputs[0].split("@")[0]}')
        return pipe

    @staticmethod
    def enforce_AT_schema_on_embedding_processors  (pipe):
        """For every embedding provider and consumer, enforce that their output col is named <output_level>@storage_ref for output_levels word,chunk,sentence aka document , i.e. word_embed@elmo or sentence_embed@elmo etc.."""
        for c in pipe.components:
            if ComponentUtils.is_embedding_provider(c):
                if '@' not in c.info.outputs[0]:
                    new_embed_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(c, 'output')
                    c.info.outputs = [new_embed_AT_ref]
                    c.info.spark_output_column_names = [new_embed_AT_ref]
                    # c.model.setOutputCol(new_embed_AT_ref[0]) # why [0] here?? bug!
                    c.model.setOutputCol(new_embed_AT_ref)

            if ComponentUtils.is_embedding_consumer(c):
                input_embed_col = ComponentUtils.extract_embed_col(c)
                if '@' not in input_embed_col:
                    # storage_ref = StorageRefUtils.extract_storage_ref(c)
                    # new_embed_col_with_AT_notation = input_embed_col+"@"+storage_ref
                    # TODO SET STORAGE REF ON TRAINABLE MODELS!!!!!!!!
                    ## OTHERWHISE ENFORCING AT STORAGE REF WILL BREAK STUFFFZ!
                    # STORAGE REF MUST MATCH UP TO THE EMBEDDINGS WE ARE FEEDING!!
                    new_embed_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(c, 'input')
                    c.info.inputs.remove(input_embed_col)
                    c.info.inputs.append(new_embed_AT_ref)
                    c.info.spark_input_column_names.remove(input_embed_col)
                    c.info.spark_input_column_names.append(new_embed_AT_ref)
                    c.model.setInputCols(c.info.inputs)






        return pipe


    @staticmethod
    def enforce_NLU_columns_to_NLP_columns  (pipe):
        """for every component, set its inputs and outputs to the ones configured on the NLU component."""
        for c in pipe.components:
            if c.info.name == 'document_assembler' : continue
            c.model.setOutputCol(c.info.outputs[0])
            c.model.setInputCols(c.info.inputs)
            c.info.spark_input_column_names = c.info.inputs
            c.info.spark_output_column_names = c.info.outputs

        return pipe
    @staticmethod
    def is_converter_component_resolution_reference(reference:str)-> bool:
        if 'chunk_emb' in reference : return True

    @staticmethod
    def configure_component_output_levels_to_sentence(pipe):
        '''
        Configure pipe components to output level document. Substitute every occurence of <document> to <sentence> for every component that feeds from <document
        :param pipe: pipe to be configured
        :return: configured pipe
        '''
        logger.info('Configuring components to sentence level')
        for c in pipe.components:
            if 'document' in c.info.inputs and 'sentence' not in c.info.inputs and 'sentence' not in c.info.outputs :
                logger.info(f"Configuring C={c.info.name}  of Type={type(c.model)}")
                c.info.inputs.remove('document')
                c.info.inputs.append('sentence')
                c.model.setInputCols(c.info.inputs)
            if 'document' in c.info.spark_input_column_names and 'sentence' not in c.info.spark_input_column_names and 'sentence' not in c.info.spark_output_column_names  :
                c.info.spark_input_column_names.remove('document')
                c.info.spark_input_column_names.append('sentence')
                if c.info.type == 'sentence_embeddings':c.info.output_level = 'sentence'
        return pipe.components

    @staticmethod
    def configure_component_output_levels_to_document(pipe):
        '''
        Configure pipe components to output level document. Substitute every occurence of <sentence> to <document> for every component that feeds from <sentence>
        :param pipe: pipe to be configured
        :return: configured pipe
        '''
        logger.info('Configuring components to document level')
        for c in pipe.components:
            if 'sentence' in c.info.inputs and 'document' not in c.info.inputs and 'document' not in c.info.outputs:
                logger.info(f"Configuring C={c.info.name}  of Type={type(c.model)}")
                c.info.inputs.remove('sentence')
                c.info.inputs.append('document')
                c.model.setInputCols(c.info.inputs)
            if 'sentence' in c.info.spark_input_column_names and 'document' not in c.info.spark_input_column_names and 'document' not in c.info.spark_output_column_names  :
                c.info.spark_input_column_names.remove('sentence')
                c.info.spark_input_column_names.append('document')
                if c.info.type == 'sentence_embeddings':c.info.output_level = 'document'
        return pipe.components

    @staticmethod
    def has_sentence_detector(pipe):
        """Check for NLUPipieline if it contains sentence detector"""
        for c in pipe.components :
            if isinstance(c.model, (SentenceDetectorDLModel, SentenceDetector, SentenceDetectorDLApproach)) : return True
        return False

    @staticmethod
    def configure_component_output_levels(pipe, force_level=''):
        '''
        This method configures sentenceEmbeddings and Classifier components to output at a specific level
        This method is called the first time .predit() is called and every time the output_level changed
        If output_level == Document, then sentence embeddings will be fed on Document col and classifiers recieve doc_embeds/doc_raw column, depending on if the classifier works with or withouth embeddings
        If output_level == sentence, then sentence embeddings will be fed on sentence col and classifiers recieve sentence_embeds/sentence_raw column, depending on if the classifier works with or withouth embeddings. IF sentence detector is missing, one will be added.
        pipe : NLU pipeline
        return : a new list of components, which columns cnfigured for output level
        '''

        from nlu import Util

        if pipe.output_level == 'sentence' or force_level=='sentence':
            if not PipeUtils.has_sentence_detector(pipe) :
                logger.info("Adding missing Sentence Detector")
                pipe.is_fitted=False
                pipe.components.insert(1,Util('deep_sentence_detector', nlu_ref='sentence'))
            return PipeUtils.configure_component_output_levels_to_sentence(pipe)
        elif pipe.output_level == 'document' or force_level=='document' :
            return PipeUtils.configure_component_output_levels_to_document(pipe)




    @staticmethod
    def check_if_component_is_in_pipe(pipe, component_name_to_check, check_strong=True):
        """Check if a component with a given name is already in a pipe """
        for c in pipe.components :
            if   check_strong and component_name_to_check == c.info.name : return True
            elif not check_strong and component_name_to_check in c.info.name : return True
        return False

    @staticmethod
    def check_if_there_component_with_col_in_components(component_list, features, except_component):
        """For a given list of features and a list of components, see if there are components taht provide this feature
        If yes, True, otherwise False
        """
        for c in component_list : 
            if c.info.outputs[0] != except_component.info.outputs[0] :
                for f in ComponentUtils.clean_irrelevant_features(c.info.spark_output_column_names, True):
                    if f in features : return True

        return False

    @staticmethod
    def is_leaf_node(c,pipe)-> bool:
        """Check if a component is a leaf in the DAG.
        We verify by checking if any other_c is feeding from c.
        If yes, it is not a leaf. If nobody feeds from c, it's a leaf.
        """
        inputs = c.info.inputs
        for other_c in pipe.components:
            if c is not other_c :
                for f in other_c.info.inputs :1


        return False



    @staticmethod
    def subsitute_leaf_output_names(pipe):
        """Change all output column names of leaves to something nicer, if they not already
        use AT notation"""
    # TODO WIP
        for c in pipe.components:
            if PipeUtils.is_leaf_node(c,pipe) and not ComponentUtils.has_AT_notation():
                # update name
                1

        return pipe



    @staticmethod
    def clean_AT_storage_refs(pipe):
        """Removes AT notation from all columns. Useful to reset pipe back to default state"""

        for c in pipe.components:
            c.info.inputs  = [f.split('@')[0] for f in c.info.inputs]
            c.info.outputs = [f.split('@')[0] for f in c.info.outputs]

            c.info.spark_input_column_names  = [f.split('@')[0] for f in c.info.spark_input_column_names]
            c.info.spark_output_column_names = [f.split('@')[0] for f in c.info.spark_output_column_names]

            c.info.spark_input_column_names  =c.info.inputs.copy()#  [f.split('@')[0] for f in c.info.spark_input_column_names]
            c.info.spark_output_column_names =c.info.outputs.copy()#  [f.split('@')[0] for f in c.info.spark_output_column_names]


        return pipe



    @staticmethod
    def rename_duplicate_cols(pipe):
        """Rename cols with duplicate names"""
        for i,c in enumerate(pipe.components):
            for other_c in pipe.components:
                if c is other_c : continue
                if  c.info.outputs[0] == other_c.info.outputs[0]:
                    c.info.outputs[0] = f'{c.info.outputs[0]}_{str(i)}'
                    c.info.spark_output_column_names[0] = f'{c.info.spark_output_column_names[0]}'

        return pipe




    @staticmethod
    def find_trainable_embed_consumer(pipe):
        for i,c in enumerate(pipe.components):
            if isinstance(c.model, NerDLApproach): return i, 'word_embeddings'
            if isinstance(c.model, (ClassifierDLApproach,SentimentDLApproach,MultiClassifierDLApproach)): return i, 'sentence_embeddings'
            if pipe.has_licensed_components:
                from sparknlp_jsl.annotator import ChunkEntityResolverApproach, SentenceEntityResolverApproach
                if isinstance(c.model, ChunkEntityResolverApproach):    return i, 'chunk_embeddings'
                if isinstance(c.model, SentenceEntityResolverApproach): return i, 'sentence_embeddings'

        return -1, None



from typing import List
import logging
logger = logging.getLogger('nlu')
import inspect
from nlu.pipe.pipe_components import SparkNLUComponent
from nlu.pipe.utils.storage_ref_utils import StorageRefUtils

class ComponentUtils():
    """Component and Column Level logic operations and utils"""
    @staticmethod
    def config_chunk_embed_converter(converter:SparkNLUComponent)->SparkNLUComponent:
        '''For a Chunk to be added to a pipeline, configure its input/output and set storage ref to amtch the storage ref and
        enfore storage ref notation. This will be used to infer backward later which component should feed this consumer'''
        storage_ref = StorageRefUtils.extract_storage_ref(converter)
        input_embed_col = ComponentUtils.extract_embed_col(converter)
        new_embed_col_with_AT_notation = input_embed_col+"@"+storage_ref
        converter.info.inputs.remove(input_embed_col)
        converter.info.inputs.append(new_embed_col_with_AT_notation)
        converter.info.spark_input_column_names.remove(input_embed_col)
        converter.info.spark_input_column_names.append(new_embed_col_with_AT_notation)
        converter.model.setInputCols(converter.info.inputs)


        return converter


    @staticmethod
    def clean_irrelevant_features(feature_list, remove_AT_notation = False):
        '''
        Remove irrelevant features from a list of component features
        Also remove the @notation from names, since they are irrelevant for ordering
        :param feature_list: list of features
        :param remove_AT_notation: remove AT notation from c names if true. Used for sorting
        :return: list with only relevant feature names
        '''
        # remove irrelevant missing features for pretrained models
        if 'text'            in feature_list: feature_list.remove('text')
        if 'raw_text'        in feature_list: feature_list.remove('raw_text')
        if 'raw_texts'       in feature_list: feature_list.remove('raw_texts')
        if 'label'           in feature_list: feature_list.remove('label')
        if 'sentiment_label' in feature_list: feature_list.remove('sentiment_label')
        if '%%%feature_elements%%%' in feature_list : feature_list.remove('%%%feature_elements%%%')
        if remove_AT_notation:
            new_cs = []
            for c in feature_list :new_cs.append(c.split("@")[0])
            return new_cs
        return feature_list


    @staticmethod
    def component_has_embeddings_requirement(component:SparkNLUComponent):
        '''
        Check for the input component, wether it depends on some embedding. Returns True if yes, otherwise False.
        :param component:  The component to check
        :return: True if the component needs some specifc embedding (i.e.glove, bert, elmo etc..). Otherwise returns False
        '''

        if type(component) == list or type(component) == set:
            for feature in component:
                if 'embed' in feature: return True
            return False
        else:
            if component.info.type == 'word_embeddings': return False
            if component.info.type == 'sentence_embeddings': return False
            if component.info.type == 'chunk_embeddings': return False
            for feature in component.info.inputs:
                if 'embed' in feature: return True
        return False
    @staticmethod
    def component_has_embeddings_provisions(component:SparkNLUComponent):
        '''
        Check for the input component, wether it depends on some embedding. Returns True if yes, otherwise False.
        :param component:  The component to check
        :return: True if the component needs some specifc embedding (i.e.glove, bert, elmo etc..). Otherwise returns False
        '''
        if type(component) == type(list) or type(component) == type(set):
            for feature in component:
                if 'embed' in feature: return True
            return False
        else:
            for feature in component.info.outputs:
                if 'embed' in feature: return True
        return False


    @staticmethod
    def extract_storage_ref_AT_notation_for_embeds(component:SparkNLUComponent, col='input'):
        '''
        Extract <col>_embed_col@storage_ref notation from a component if it has a storage ref, otherwise '

        :param component:  To extract notation from
        :cols component:  Wether to extract for the input or output col
        :return: '' if no storage_ref, <col>_embed_col@storage_ref otherwise
        '''
        # if not StorageRefUtils.has_storage_ref(component) :
        #     if   col =='input' : return component.info.inputs
        #     elif col =='output': return component.info.outputs
        if   col =='input'  : e_col    = next(filter(lambda s : 'embed' in s, component.info.inputs))
        elif col =='output' : e_col    = next(filter(lambda s : 'embed' in s, component.info.outputs))

        stor_ref = StorageRefUtils.extract_storage_ref(component)
        return e_col + '@' + stor_ref

    @staticmethod
    def is_embedding_provider(component:SparkNLUComponent) -> bool:
        """Check if a NLU Component returns embeddings """
        if 'embed' in component.info.outputs[0]:
            return True
        else:
            return False

    @staticmethod
    def is_embedding_consumer(component:SparkNLUComponent) -> bool:
        """Check if a NLU Component consumes embeddings """
        return any('embed' in i for i in component.info.inputs)


    @staticmethod
    def is_embedding_converter  (component : SparkNLUComponent) -> bool:
        """Check if NLU component is embedding converter """
        if component.info.name in ['chunk_embedding_converter', 'sentence_embedding_converter']: return True
        return False


    @staticmethod
    def is_NER_provider(component:SparkNLUComponent) -> bool:
        """Check if a NLU Component wraps a NER/NER-Medical model """
        return component.info.name.split('@')[0] in ['named_entity_recognizer_dl', 'medical_named_entity_recognizer_dl','ner', 'named_entity_recognizer_dl_healthcare', 'med_ner']

    @staticmethod
    def is_NER_converter(component:SparkNLUComponent) -> bool:
        """Check if a NLU Component wraps a NER-IOB to NER-Pr etty converter """
        return component.info.name in ['NerToChunkConverter','ner_to_chunk_converter','ner_to_chunk_converter_licensed']

    @staticmethod
    def extract_NER_col(component:SparkNLUComponent, column='input')->str:
        """Extract the exact name of the NER column in the component"""
        if column == 'input':
            for f in component.info.spark_input_column_names:
                if 'ner' in f : return f
        if column == 'output':
            for f in component.info.spark_output_column_names:
                if 'ner' in f : return f

    @staticmethod
    def extract_NER_converter_col(component:SparkNLUComponent, column='input')->str:
        """Extract the exact name of the NER-converter  column in the component"""
        if column == 'input':
            for f in component.info.spark_input_column_names:
                if 'ner' in f : return f
        if column == 'output':
            for f in component.info.spark_output_column_names:
                if 'entities' in f : return f



    @staticmethod
    def extract_embed_col(component:SparkNLUComponent, column='input')->str:
        """Extract the exact name of the embed column in the component"""
        if column == 'input':
            for c in component.info.spark_input_column_names:
                if 'embed' in c : return c
        if column == 'output':
            for c in component.info.spark_output_column_names:
                if 'embed' in c : return c

    @staticmethod
    def is_untrained_model(component:SparkNLUComponent)->bool:
        '''
        Check for a given component if it is an embelishment of an traianble model.
        In this case we will ignore embeddings requirements further down the logic pipeline
        :param component: Component to check
        :return: True if it is trainable, False if not
        '''
        if 'is_untrained' in dict(inspect.getmembers(component.info)).keys(): return True
        return False



    @staticmethod
    def set_storage_ref_attribute_of_embedding_converters(pipe_list):
        """For every embedding converter, we set storage ref attr on it, based on what the storage ref from it's provider is """
        for converter in pipe_list:
            if ComponentUtils.is_embedding_provider(converter) and  ComponentUtils.is_embedding_converter(converter) :
                embed_col = ComponentUtils.extract_embed_col(converter)
                for provider in pipe_list:
                    if embed_col in provider.info.outputs:
                        converter.info.storage_ref = StorageRefUtils.extract_storage_ref(provider)
                # if c.info.name =='ChunkEmbeddings' : c.model.setOutputCol(level_AT_ref[0])
                # else : c.model.setOutputCol(level_AT_ref)
        return pipe_list

    @staticmethod
    def extract_embed_level_identity(component, col='input'):
        """Figure out if component feeds on chunk/sent aka doc/word emb for either nput or output cols"""
        if col =='input':
            if any(filter(lambda s : 'document_embed' in s , component.info.inputs)): return 'document_embeddings'
            if any(filter(lambda s : 'sentence_embed' in s , component.info.inputs)): return 'sentence_embeddings'
            if any(filter(lambda s : 'chunk_embed' in s , component.info.inputs)): return 'chunk_embeddings'
            if any(filter(lambda s : 'token_embed' in s , component.info.inputs)): return 'token_embeddings'
        elif col == 'output':
            if any(filter(lambda s : 'document_embed' in s , component.info.outputs)): return 'document_embeddings'
            if any(filter(lambda s : 'sentence_embed' in s , component.info.outputs)): return 'sentence_embeddings'
            if any(filter(lambda s : 'chunk_embed' in s , component.info.outputs)): return 'chunk_embeddings'
            if any(filter(lambda s : 'token_embed' in s , component.info.outputs)): return 'token_embeddings'

    @staticmethod
    def are_producer_consumer_matches(e_consumer:SparkNLUComponent,e_provider:SparkNLUComponent) -> bool:
        """Check for embedding_consumer and embedding_producer if they match storage_ref and output level wise wise """
        if StorageRefUtils.extract_storage_ref(e_consumer)== StorageRefUtils.extract_storage_ref(e_provider):
            if ComponentUtils.extract_embed_level_identity(e_consumer, 'input') ==  ComponentUtils.extract_embed_level_identity(e_provider, 'output'):
                return True
            ## TODO FALL BACK FOR BAD MATCHES WHICH ACTUALLY MATCH-> consult name space
        return False


    @staticmethod
    def get_nlu_ref_identifier(component:SparkNLUComponent) -> str:
        """The tail of a NLU ref after splitting on '.' gives a unique identifier for NON-Aliased components
         If result is '' , model UID will be used as identifier
         """
        tail = ''

        if hasattr(component.info, 'nlu_ref') and component.info.nlu_ref !='':
            tail = component.info.nlu_ref.split('.')[-1].split('@')[-1]
        if tail == '' :
            logger.warning(f"Could not deduct tail from component={component}. This is intended for CustomModelComponents used in offline mode")
            tail = str(component.model)
        return tail

    @staticmethod
    def remove_storage_ref_from_features(features:List[str]):
        """Clean storage ref from every str in list """
        return [f.split('@')[0] for f in features]



from sparknlp_jsl.annotator import AssertionLogRegModel, AssertionLogRegApproach

class AssertionLogReg:
    @staticmethod
    def get_default_model():
        return AssertionLogRegModel.pretrained() \
                               .setInputCols(["sentence", "entities", "word_embeddings"]) \
                               .setOutputCol("assertion")


    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        return AssertionLogRegModel.pretrained(name, language,bucket) \
            .setInputCols(["sentence", "entities", "word_embeddings"]) \
            .setOutputCol("assertion")

    def get_default_trainable_model(self):
        return AssertionLogRegApproach()\
            .setInputCols(["sentence", "entities", "word_embeddings"]) \
            .setOutputCol("assertion")\
            .setLabelCol('label')



from nlu.pipe.pipe_components import SparkNLUComponent


class Chunker(SparkNLUComponent):

    def __init__(self, annotator_class='default_chunker', language='en', component_type='chunker', get_default = True, nlp_ref='', nlu_ref='',  model=None, lang='en',loaded_from_pretrained_pipe=False,is_licensed=False):
        if model != None : self.model = model
        else : 
            if annotator_class == 'default_chunker':
                from nlu import DefaultChunker
                if get_default : self.model =  DefaultChunker.get_default_model()
                else : self.model =  DefaultChunker.get_default_model()  # there are no pretrained chunkers, only default 1
            if annotator_class == 'ngram':
                from nlu import NGram
                if get_default : self.model =  NGram.get_default_model()
                else : self.model =  NGram.get_default_model()  # there are no pretrained chunkers, only default 1
            if annotator_class == 'contextual_parser':
                from nlu.components.chunkers.contextual_parser.contextual_parser import ContextualParser
                if get_default : self.model =  ContextualParser.get_default_model()
                else: self.model =  ContextualParser.get_default_model()


        SparkNLUComponent.__init__(self, annotator_class, component_type, nlu_ref, nlp_ref, lang,loaded_from_pretrained_pipe , is_licensed)
